#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <sys/select.h>
#include <glob.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
#include <pthread.h>
#include <sys/inotify.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>


const char program[] = "logOSM";
const char version[] = "0.1";
const char verdate[] = "2014 Mar 27";

const int MaxLineLength=2000;

const char DefaultProgram[] = "/export/home/cc/vlba/bin/observMon.sh";
const char DefaultWatchFile[] = "/var/spool/cc/logOSM.watch";
const int DefaultFilesPerDay = 1;
const int DefaultUserId = 5854;	/* vlbamon */

pid_t executorpid = 0;
int die = 0;
typedef void (*sighandler_t)(int);
sighandler_t oldsiginthand;

void siginthand(int j)
{
	int status; // unused status

	fprintf(stderr, "Sending term signal to pid %d\n", executorpid);

	die = 1;
}

int usage(int argc, char **argv)
{
	printf("\n%s version %s\n\n", program, version);
	printf("Matthias Bark %s\n\n", verdate);
	printf("A program to start VLBA OSM and log its output.\n\n");
	printf("Usage : %s <time> \n\n", argv[0]);
	printf("<time> is a time delay in seconds before starting the program.\n");
	printf("observMon.sh is started and debug data is logged to\n");
	printf("/export/home/cc/vlba/logs/<year>_<doy>_<time>.osm.log\n\n");
	
	return 0;
}


pid_t startOSM(int *infp, int *outfp, int *errfp, const char *program, int verbose)
{
        int status;
	int p_stdin[2], p_stdout[2], p_stderr[2];
	pid_t pid;
	
	if(pipe(p_stdin) != 0 || pipe(p_stdout) != 0 || pipe(p_stderr) != 0)
	{
		return -1;
	}

	pid = fork();

        // error
	if(pid < 0)
	{
		return pid;
	}
        // child process
	else if(pid == 0)
	{
		close(p_stdin[1]);
		dup2(p_stdin[0], 0);
		close(p_stdout[0]);
		dup2(p_stdout[1], 1);
		close(p_stderr[0]);
		dup2(p_stderr[1], 2);

		printf("Trying to start <%s>\n", program);
		execl(program, (char *)NULL);
		perror("execl");
		fprintf(stderr, "Error starting %s\n", program);
		// if there is an error make sure the script starts with #!/bin/bash

		exit(1);
	}
        // parent process - it's ok to wait as the script we are calling exits
        wait(&status);

	if(infp == NULL)
	{
		close(p_stdin[1]);
	}
	else
	{
		*infp = p_stdin[1];
	}

	if(outfp == NULL)
	{
		close(p_stdout[0]);
	}
	else
	{
		*outfp = p_stdout[0];
	}

	if(errfp == NULL)
	{
		close(p_stderr[0]);
	}
	else
	{
		*errfp = p_stderr[0];
	}

	return pid;
}

void *gzipfiles(void *data)
{
	const char match[] = "/export/home/cc/vlba/logs/????_???_??_??_??.osm.log";
	char cmd[256];
	glob_t globbuf;
	int i, n, v;
	const char *except = (const char *)data;

	glob(match, 0, 0, &globbuf);

	n = globbuf.gl_pathc;
	
	for(i = 0; i < n; i++)
	{
		if(strcmp(except, globbuf.gl_pathv[i]) == 0)
		{
			continue;
		}
		sprintf(cmd, "gzip -f %s", globbuf.gl_pathv[i]);
		v = system(cmd);
	}

	globfree(&globbuf);
	free(data);

	pthread_exit(0);
}

FILE *newfile(FILE *old, time_t t, int filesPerDay, int verbose)
{
	FILE *out;
	char *filename;
	struct tm tmg;
	int fileOfDay;
	pthread_t pt;

	fileOfDay = (t % 86400)*filesPerDay/86400;
	
	/* adjust time to correspond to start of a day segment */
	t = 86400*(t/86400) + 86400*fileOfDay/filesPerDay;

	gmtime_r(&t, &tmg);

	if(old)
	{
		fclose(old);
	}
	
	filename = (char *)malloc(256);
	sprintf(filename, "/export/home/cc/vlba/logs/%04d_%03d_%02d_%02d_%02d.osm.log",
		tmg.tm_year+1900, tmg.tm_yday+1, tmg.tm_hour, tmg.tm_min, tmg.tm_sec);
	if(verbose)
	{
		printf("Opening %s day fraction = %d/%d\n", filename, fileOfDay, filesPerDay);
	}

	pthread_create(&pt, 0, gzipfiles, filename);
	pthread_detach(pt);

	out = fopen(filename, "a");

	return out;
}

/* return value: == 1 -> run, >= 0 -> stop */
int getOSMState(const char *watchFile, int init, char *comment)
{
	FILE *in, *out;
	char line[MaxLineLength];
	char *v;
	int state = 0;

	in = fopen(watchFile, "r");

	if(in)
	{
		v = fgets(line, MaxLineLength, in);
		if(v)
		{
			state = atoi(line);
			if(comment)
			{
				v = fgets(comment, MaxLineLength, in);
			}
		}
		fclose(in);
	}
	else
	{
		if(init)
		{
			out = fopen(watchFile, "w");
			fprintf(out, "1\nInitialized by %s\n", program);
			fclose(out);
		}
	}

	return state;
}

/* fds[] are: [stdout, stderr, inotify] */
void watchpipe(int fds[3], int filesPerDay, const char *watchFile, int verbose)
{
	char str[MaxLineLength+1];
	char comment[MaxLineLength];
	time_t lastt=0;
	time_t t;
	struct tm tmg;
	FILE *out=0;
	int l;
	fd_set readset;
	int maxfd;
	int i, v, state;
	struct timeval tv;
	int status;

	maxfd = -1;
	for(i = 0; i < 3; i++)
	{
		if(fds[i] > maxfd)
		{
			maxfd = fds[i];
		}
	}
	maxfd++;	/* needs to be 1 greater than max actually */

	while(!die)
	{
		FD_ZERO(&readset);
		FD_SET(fds[0], &readset);
		FD_SET(fds[1], &readset);
		FD_SET(fds[2], &readset);
		tv.tv_sec = 1;
		tv.tv_usec = 0;

		v = select(maxfd, &readset, 0, 0, &tv);

		if(v < 0)
		{
			if(!die)
			{
				printf("Error: select failed\n");
			}

			return;
		}

		if(v == 0)
		{
			/* timed out */

			continue;
		}

		t = time(0);
		gmtime_r(&t, &tmg);
		if(t != lastt)
		{
			if((long long)t*(long long)filesPerDay/86400LL != (long long)lastt*(long long)filesPerDay/86400LL)
			{
				out = newfile(out, t, filesPerDay, verbose);
			}
			lastt = t;
			fprintf(out, "\n");
			fprintf(out, "%04d/%03d %02d:%02d:%02d\n",
				tmg.tm_year+1900, tmg.tm_yday+1,
				tmg.tm_hour, tmg.tm_min, tmg.tm_sec);
		}
	
		/* handle stdout and stderr reads */
		for(i = 0; i < 2; i++)
		{
			if(!FD_ISSET(fds[i], &readset))
			{
				continue;
			}
			l = read(fds[i], str, MaxLineLength);
			if(l < 0)
			{
				break;
			}
			str[l] = 0;

			fprintf(out, "%s", str);
		}

		/* handle inotify read */
		if(FD_ISSET(fds[2], &readset))
		{
			const struct inotify_event *ie;
			l = read(fds[i], str, MaxLineLength);
			if(l < 0)
			{
				break;
			}
			str[l] = 0;
			ie = (const struct inotify_event *)str;

			state = getOSMState(watchFile, 0, comment);
			fprintf(out, "inotify_event: %d %u state=%d comment=%s\n", ie->wd, ie->mask, state, comment);
			printf(" inotify_event: %d %u state=%d comment=%s\n", ie->wd, ie->mask, state, comment);
			if(state <= 0)
			{
				char bbuf[20];
				int killPID = -1;
				FD_ZERO(&readset);
				FILE *f = popen("ps ax | grep -v grep | grep ObservScriptMonitor | awk '{print $1}'", "r");
				fread(bbuf, 20, 1, f);
				killPID = atoi(bbuf);
				if( killPID == 0 || killPID == -1 ) {
					printf("Error - something failed when trying to read OSM PID\n");
				} else {
					kill(atoi(bbuf), SIGTERM);
				}
				if(state < 0)
				{
					die = 1;
				}
				break;
			}
		}
		fflush(out);
	}
	fclose(out);
	sleep(1);
}

int main(int argc, char **argv)
{
	int t;
	int infd=0, fds[3];
	int v;
	int wd;
	int filesPerDay = DefaultFilesPerDay;
	int userId = DefaultUserId;
	int verbose = 0;
	const char *watchFile = DefaultWatchFile;
	const char *program = DefaultProgram;
	char line[MaxLineLength];
	int status;

	if(argc > 1)
	{
		if(strcmp(argv[1], "-h") == 0)
		{
			return usage(argc, argv);
		}
		else if(strcmp(argv[1], "-t") == 0)
		{
			userId = -1;
			filesPerDay = 2880;
			watchFile = "/tmp/logOSM.watch";
			verbose = 1;
		}
		else
		{
			t = atoi(argv[1]);
			sleep(t);
		}
	}

	if(userId >= 0)
	{
		setuid(userId);
	}

	getOSMState(watchFile, 1, 0);
	fds[2] = inotify_init();
	wd = inotify_add_watch(fds[2], watchFile, IN_CLOSE_WRITE | IN_DELETE_SELF);
	printf("Watch descriptor = %d\n", wd);

	while(1)
	{
		status = getOSMState(watchFile, 0, 0);
		if(status == -1)
		{
			break;
		}
		if(status)
		{
			fds[0] = fds[1] = 0;
			
			oldsiginthand = signal(SIGINT, siginthand);
			executorpid = startOSM(&infd, fds+0, fds+1, program, verbose);
			
			if(executorpid == 0)
			{
				fprintf(stderr, "Error starting %s\n", program);
			}
			else
			{
				printf("OSM pid is %d\n", executorpid);
				watchpipe(fds, filesPerDay, watchFile, verbose);
				close(infd);
				close(fds[0]);
				close(fds[1]);
			}

			signal(SIGINT, oldsiginthand);
		}
		if(die)
		{
			break;
		}

		/* here, safely assume that getting here implied a file change on watchFile*/
		v = read(fds[2], line, MaxLineLength-1);
	}

	inotify_rm_watch(fds[2], wd);

	return 0;
}
